"""Utility functions related to parsing of safetensors metadata of models."""

import logging

from huggingface_hub import get_safetensors_metadata
from huggingface_hub.errors import NotASafetensorsRepoError

from .logging_utils import log_once
from .utils import get_hf_token


def get_num_params_from_safetensors_metadata(
    model_id: str, revision: str, api_key: str | None
) -> int | None:
    """Get the number of parameters from the safetensors metadata.

    Args:
        model_id:
            The model ID.
        revision:
            The revision of the model.
        api_key:
            The API key to use for authentication with the Hugging Face Hub. Can be
            None if no API key is needed.

    Returns:
        The number of parameters, or None if the metadata could not be found.
    """
    try:
        metadata = get_safetensors_metadata(
            repo_id=model_id, revision=revision, token=get_hf_token(api_key=api_key)
        )
    except NotASafetensorsRepoError:
        log_once(
            "The number of parameters could not be determined for the model "
            f"{model_id}, since the model is not stored in the safetensors format. "
            "If this is your own model, then you can use this Hugging Face Space to "
            "convert your model to the safetensors format: "
            "https://huggingface.co/spaces/safetensors/convert.",
            level=logging.WARNING,
        )
        return None

    parameter_count_dict = metadata.parameter_count
    match len(parameter_count_dict):
        case 0:
            log_once(
                "Failed to determine the number of parameters for the model "
                f"{model_id}, even though the model is stored in the safetensors "
                "format. Please report this issue at "
                "https://github.com/EuroEval/EuroEval/issues.",
                level=logging.WARNING,
            )
            return None
        case 1:
            return max(parameter_count_dict.values())
        case _:
            log_once(
                f"The model {model_id} has multiple parameter count entries in its "
                f"safetensors metadata: {parameter_count_dict}. Using the largest one.",
                level=logging.DEBUG,
            )
            return max(parameter_count_dict.values())



"""Data models used in EuroEval."""

import collections.abc as c
import importlib.metadata
import json
import logging
import re
import typing as t
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path

import pydantic
import torch
from transformers.generation.configuration_utils import GenerationConfig

from .constants import ATTENTION_BACKENDS, MAX_NUMBER_OF_LOGGING_LANGUAGES
from .enums import Device, GenerativeType, ModelType, TaskGroup
from .exceptions import InvalidBenchmark
from .languages import (
    ENGLISH,
    EUROPEAN_PORTUGUESE,
    NORWEGIAN,
    NORWEGIAN_BOKMÅL,
    NORWEGIAN_NYNORSK,
    PORTUGUESE,
    Language,
)
from .logging_utils import log_once
from .metrics.base import Metric
from .types import ScoreDict

if t.TYPE_CHECKING:
    from .enums import InferenceBackend


def get_package_version(package_name: str) -> str | None:
    """Get the version of a package.

    Args:
        package_name:
            The name of the package.

    Returns:
        The version of the package, or None if the package is not installed.
    """
    try:
        return importlib.metadata.version(package_name)
    except importlib.metadata.PackageNotFoundError:
        return None


@dataclass
class PromptConfig:
    """Configuration for task-specific prompting across languages.

    Defines the prompt templates needed for evaluating a specific task in a given
    language.

    Attributes:
        default_prompt_prefix:
            The default prefix to use in the few-shot prompt.
        default_prompt_template:
            The default template for the prompt to use when benchmarking the dataset
            using few-shot evaluation.
        default_instruction_prompt:
            The default prompt to use when benchmarking the dataset using
            instruction-based evaluation.
        default_prompt_label_mapping:
            The default mapping from the labels to another phrase which is used as a
            substitute for the label in few-shot evaluation. If set to "auto", the
            mapping will be set to a 1:1 mapping between the labels and themselves.
    """

    default_prompt_prefix: str
    default_prompt_template: str
    default_instruction_prompt: str
    default_prompt_label_mapping: dict[str, str] | t.Literal["auto"]


@dataclass
class Task:
    """A dataset task.

    Attributes:
        name:
            The name of the task.
        task_group:
            The task group of the task.
        template_dict:
            The template dictionary for the task, from language (or language tuples) to
            prompt template.
        metrics:
            The metrics used to evaluate the task.
        default_num_few_shot_examples:
            The default number of examples to use when benchmarking the task using
            few-shot evaluation. For a classification task, these will be drawn evenly
            from each label.
        default_max_generated_tokens:
            The default maximum number of tokens to generate when benchmarking the task
            using few-shot evaluation.
        default_labels (optional):
            The default labels for datasets using this task. Can be None if the labels
            should be set manually in the dataset configs. Defaults to an empty tuple.
        requires_zero_shot (optional):
            Whether to only allow zero-shot evaluation for this task. If True, the
            task will not be evaluated using few-shot examples.
        uses_structured_output (optional):
            Whether the task uses structured output. If True, the task will return
            structured output (e.g., BIO tags for NER). Defaults to False.
        uses_logprobs (optional):
            Whether the task uses log probabilities. If True, the task will return
            log probabilities for the generated tokens. Defaults to False.
        requires_logprobs (optional):
            Whether the task requires log probabilities. Implies `uses_logprobs`.
        default_allowed_model_types (optional):
            A list of model types that are allowed to be evaluated on this task.
            Defaults to all model types being allowed.
        default_allowed_generative_types (optional):
            A list of generative model types that are allowed to be evaluated on this
            task. If None, all generative model types are allowed. Only relevant if
            `allowed_model_types` includes generative models.
        default_allow_invalid_model_outputs (optional):
            Whether to allow invalid model outputs. This is only relevant for generative
            models on classification tasks, where the model may generate an output
            which is not one of the allowed labels. If True, the model output will be
            mapped to the closest valid label. If False, the model output will be
            considered incorrect and the evaluation will be aborted. Defaults to True.
    """

    model_config = pydantic.ConfigDict(
        protected_namespaces=(), arbitrary_types_allowed=True
    )

    name: str
    task_group: TaskGroup
    template_dict: (
        dict[Language, PromptConfig] | dict[tuple[Language, Language], PromptConfig]
    )
    metrics: c.Sequence[Metric]
    default_num_few_shot_examples: int
    default_max_generated_tokens: int
    default_labels: c.Sequence[str] | None = tuple()
    requires_zero_shot: bool = False
    uses_structured_output: bool = False
    uses_logprobs: bool = False
    requires_logprobs: bool = False
    default_allowed_model_types: c.Sequence[ModelType] = field(
        default_factory=lambda: [ModelType.ENCODER, ModelType.GENERATIVE]
    )
    default_allowed_generative_types: c.Sequence[GenerativeType] = field(
        default_factory=lambda: [
            GenerativeType.BASE,
            GenerativeType.INSTRUCTION_TUNED,
            GenerativeType.REASONING,
        ]
    )
    default_allow_invalid_model_outputs: bool = True

    def __post_init__(self) -> None:
        """Post-initialisation checks."""
        self.uses_logprobs = self.uses_logprobs or self.requires_logprobs

    def __hash__(self) -> int:
        """Return a hash of the task."""
        return hash(self.name)


class DatasetConfig:
    """Configuration for a dataset."""

    def __init__(
        self,
        task: Task,
        languages: c.Sequence[Language],
        name: str | None = None,
        pretty_name: str | None = None,
        source: str | dict[str, str] | None = None,
        prompt_prefix: str | None = None,
        prompt_template: str | None = None,
        instruction_prompt: str | None = None,
        num_few_shot_examples: int | None = None,
        max_generated_tokens: int | None = None,
        labels: c.Sequence[str] | None = None,
        prompt_label_mapping: dict[str, str] | t.Literal["auto"] | None = None,
        allowed_model_types: c.Sequence[ModelType] | None = None,
        allowed_generative_types: c.Sequence[GenerativeType] | None = None,
        allow_invalid_model_outputs: bool | None = None,
        train_split: str | None = "train",
        val_split: str | None = "val",
        test_split: str = "test",
        bootstrap_samples: bool = True,
        unofficial: bool = False,
        _prompt_prefix: str | None = None,
        _prompt_template: str | None = None,
        _instruction_prompt: str | None = None,
        _num_few_shot_examples: int | None = None,
        _max_generated_tokens: int | None = None,
        _labels: c.Sequence[str] | None = None,
        _prompt_label_mapping: dict[str, str] | t.Literal["auto"] | None = None,
        _allowed_model_types: c.Sequence[ModelType] | None = None,
        _allowed_generative_types: c.Sequence[GenerativeType] | None = None,
        _allow_invalid_model_outputs: bool | None = None,
        _logging_string: str | None = None,
    ) -> None:
        """Initialise a DatasetConfig object.

        Args:
            task:
                The task of the dataset.
            languages:
                The ISO 639-1 language codes of the entries in the dataset.
            name (optional):
                The name of the dataset. Must be lower case with no spaces. Can be None
                if and only if the dataset config resides directly in the Hugging Face
                dataset repo. Defaults to None.
            pretty_name (optional):
                A longer prettier name for the dataset, which allows cases and spaces.
                Used for logging. Can be None if and only if the dataset config resides
                directly in the Hugging Face dataset repo. Defaults to None.
            source (optional):
                The source of the dataset, which can be a Hugging Face ID or a
                dictionary with keys "train", "val" and "test" mapping to local CSV file
                paths. Can be None if and only if the dataset config resides directly in
                the Hugging Face dataset repo. Defaults to None.
            prompt_prefix (optional):
                The prefix to use in the few-shot prompt. Defaults to the template for
                the task and language.
            prompt_template (optional):
                The template for the prompt to use when benchmarking the dataset using
                few-shot evaluation. Defaults to the template for the task and language.
            instruction_prompt (optional):
                The prompt to use when benchmarking the dataset using instruction-based
                evaluation. Defaults to the template for the task and language.
            num_few_shot_examples (optional):
                The number of examples to use when benchmarking the dataset using
                few-shot evaluation. For a classification task, these will be drawn
                evenly from each label. Defaults to the template for the task and
                language.
            max_generated_tokens (optional):
                The maximum number of tokens to generate when benchmarking the dataset
                using few-shot evaluation. Defaults to the template for the task and
                language.
            labels (optional):
                The labels in the dataset. Defaults to the template for the task and
                language.
            prompt_label_mapping (optional):
                A mapping from the labels to another phrase which is used as a
                substitute for the label in few-shot evaluation. If "auto" then the
                mapping will be set to a 1:1 mapping between the labels and themselves.
                If None then the mapping will be set to the default mapping for the task
                and language. Defaults to None.
            allowed_model_types (optional):
                A list of model types that are allowed to be evaluated on this dataset.
                Defaults to the one for the task.
            allowed_generative_types (optional):
                A list of generative model types that are allowed to be evaluated on
                this dataset. If None, all generative model types are allowed. Only
                relevant if `allowed_model_types` includes generative models. Defaults
                to the one for the task.
            allow_invalid_model_outputs (optional):
                Whether to allow invalid model outputs. This is only relevant for
                generative models on classification tasks, where the model may generate
                an output which is not one of the allowed labels. If True, the model
                output will be mapped to the closest valid label. If False, the model
                output will be considered incorrect and the evaluation will be aborted.
                Defaults to the one for the task.
            train_split (optional):
                The name of the split to use as the training set. Can be None if there
                is no training split in the dataset. Defaults to "train".
            val_split (optional):
                The name of the split to use as the validation set. Can be None if there
                is no validation split in the dataset. Defaults to "val".
            test_split (optional):
                The name of the split to use as the test set. Defaults to "test".
            bootstrap_samples (optional):
                Whether to bootstrap the dataset samples. Defaults to True.
            unofficial (optional):
                Whether the dataset is unofficial. Defaults to False.
            _prompt_prefix (optional):
                This argument is deprecated. Please use `prompt_prefix` instead.
            _prompt_template (optional):
                This argument is deprecated. Please use `prompt_template` instead.
            _instruction_prompt (optional):
                This argument is deprecated. Please use `instruction_prompt` instead.
            _num_few_shot_examples (optional):
                This argument is deprecated. Please use `num_few_shot_examples` instead.
            _max_generated_tokens (optional):
                This argument is deprecated. Please use `max_generated_tokens` instead.
            _labels (optional):
                This argument is deprecated. Please use `labels` instead.
            _prompt_label_mapping (optional):
                This argument is deprecated. Please use `prompt_label_mapping` instead.
            _allowed_model_types (optional):
                This argument is deprecated. Please use `allowed_model_types` instead.
            _allowed_generative_types (optional):
                This argument is deprecated. Please use `allowed_generative_types`
                instead.
            _allow_invalid_model_outputs (optional):
                This argument is deprecated. Please use `allow_invalid_model_outputs`
                instead.
            _logging_string (optional):
                This argument is deprecated. Please use `logging_string` instead.
        """
        # Deprecation warnings
        if _prompt_prefix is not None:
            log_once(
                "The `_prompt_prefix` argument is deprecated. Please use "
                "`prompt_prefix` instead.",
                level=logging.WARNING,
            )
            prompt_prefix = _prompt_prefix
        if _prompt_template is not None:
            log_once(
                "The `_prompt_template` argument is deprecated. Please use "
                "`prompt_template` instead.",
                level=logging.WARNING,
            )
            prompt_template = _prompt_template
        if _instruction_prompt is not None:
            log_once(
                "The `_instruction_prompt` argument is deprecated. Please use "
                "`instruction_prompt` instead.",
                level=logging.WARNING,
            )
            instruction_prompt = _instruction_prompt
        if _num_few_shot_examples is not None:
            log_once(
                "The `_num_few_shot_examples` argument is deprecated. Please use "
                "`num_few_shot_examples` instead.",
                level=logging.WARNING,
            )
            num_few_shot_examples = _num_few_shot_examples
        if _max_generated_tokens is not None:
            log_once(
                "The `_max_generated_tokens` argument is deprecated. Please use "
                "`max_generated_tokens` instead.",
                level=logging.WARNING,
            )
            max_generated_tokens = _max_generated_tokens
        if _labels is not None:
            log_once(
                "The `_labels` argument is deprecated. Please use `labels` instead.",
                level=logging.WARNING,
            )
            labels = _labels
        if _prompt_label_mapping is not None:
            log_once(
                "The `_prompt_label_mapping` argument is deprecated. Please use "
                "`prompt_label_mapping` instead.",
                level=logging.WARNING,
            )
            prompt_label_mapping = _prompt_label_mapping
        if _allowed_model_types is not None:
            log_once(
                "The `_allowed_model_types` argument is deprecated. Please use "
                "`allowed_model_types` instead.",
                level=logging.WARNING,
            )
            allowed_model_types = _allowed_model_types
        if _allowed_generative_types is not None:
            log_once(
                "The `_allowed_generative_types` argument is deprecated. Please use "
                "`allowed_generative_types` instead.",
                level=logging.WARNING,
            )
            allowed_generative_types = _allowed_generative_types
        if _allow_invalid_model_outputs is not None:
            log_once(
                "The `_allow_invalid_model_outputs` argument is deprecated. Please use "
                "`allow_invalid_model_outputs` instead.",
                level=logging.WARNING,
            )
            allow_invalid_model_outputs = _allow_invalid_model_outputs
        if _logging_string is not None:
            log_once(
                "The `_logging_string` argument is deprecated and is not used anymore. "
                "Using it will have no effect.",
                level=logging.WARNING,
            )

        self._name = name
        self._pretty_name = pretty_name
        self._source = source
        self.task = task
        self.languages = languages

        template = self.task.template_dict.get(  # pyrefly: ignore[no-matching-overload]
            self.main_language
        )
        self.prompt_prefix = (
            prompt_prefix
            if prompt_prefix is not None
            else template.default_prompt_prefix
            if template is not None
            else ""
        )
        self.prompt_template = (
            prompt_template
            if prompt_template is not None
            else template.default_prompt_template
            if template is not None
            else ""
        )
        self.instruction_prompt = (
            instruction_prompt
            if instruction_prompt is not None
            else template.default_instruction_prompt
            if template is not None
            else ""
        )
        self.num_few_shot_examples = (
            num_few_shot_examples
            if num_few_shot_examples is not None
            else self.task.default_num_few_shot_examples
        )
        self.max_generated_tokens = (
            max_generated_tokens
            if max_generated_tokens is not None
            else self.task.default_max_generated_tokens
        )
        self.labels = (
            labels if labels is not None else self.task.default_labels or list()
        )
        if prompt_label_mapping is None:
            prompt_label_mapping = (
                template.default_prompt_label_mapping
                if template is not None
                else dict()
            )
        self.prompt_label_mapping = (
            {label: label for label in self.labels}
            if prompt_label_mapping == "auto"
            else prompt_label_mapping
        )
        self.allowed_model_types = (
            allowed_model_types
            if allowed_model_types is not None
            else self.task.default_allowed_model_types
        )
        self.allowed_generative_types = (
            allowed_generative_types
            if allowed_generative_types is not None
            else self.task.default_allowed_generative_types
        )
        self.allow_invalid_model_outputs = (
            allow_invalid_model_outputs
            if allow_invalid_model_outputs is not None
            else self.task.default_allow_invalid_model_outputs
        )
        self.train_split = train_split
        self.val_split = val_split
        self.test_split = test_split
        self.bootstrap_samples = bootstrap_samples
        self.unofficial = unofficial

    @property
    def name(self) -> str:
        """The name of the dataset.

        Returns:
            The name of the dataset.

        Raises:
            ValueError:
                If the name of the dataset is not set.
        """
        if self._name is None:
            raise ValueError("The name of the dataset is not set!")
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """Set the name of the dataset.

        Args:
            value:
                The new name of the dataset.
        """
        self._name = value

    @property
    def pretty_name(self) -> str:
        """The pretty name of the dataset.

        Returns:
            The pretty name of the dataset.

        Raises:
            ValueError:
                If the pretty name of the dataset is not set.
        """
        if self._pretty_name is None:
            raise ValueError("The pretty name of the dataset is not set!")
        return self._pretty_name

    @pretty_name.setter
    def pretty_name(self, value: str) -> None:
        """Set the pretty name of the dataset.

        Args:
            value:
                The new pretty name of the dataset.
        """
        self._pretty_name = value

    @property
    def source(self) -> str | dict[str, str]:
        """The source of the dataset.

        Returns:
            The source of the dataset.

        Raises:
            ValueError:
                If the source of the dataset is not set.
        """
        if self._source is None:
            raise ValueError("The source of the dataset is not set!")
        return self._source

    @source.setter
    def source(self, value: str | dict[str, str]) -> None:
        """Set the source of the dataset.

        Args:
            value:
                The new source of the dataset.
        """
        self._source = value

    @property
    def logging_string(self) -> str:
        """The string used to describe evaluation on the dataset in logging.

        Returns:
            The logging string.
        """
        truncated_str = (
            "truncated version of the "
            if isinstance(self.source, str) and self.source.endswith("-mini")
            else ""
        )

        logging_languages = list(deepcopy(self.languages))
        if len(self.languages) > 1:
            if (
                NORWEGIAN_BOKMÅL in self.languages
                and NORWEGIAN_NYNORSK in self.languages
                and NORWEGIAN in self.languages
            ):
                logging_languages.remove(NORWEGIAN_BOKMÅL)
                logging_languages.remove(NORWEGIAN_NYNORSK)
            elif (
                NORWEGIAN_BOKMÅL in self.languages
                or NORWEGIAN_NYNORSK in self.languages
            ) and NORWEGIAN in self.languages:
                logging_languages.remove(NORWEGIAN)
            if PORTUGUESE in self.languages and EUROPEAN_PORTUGUESE in self.languages:
                logging_languages.remove(EUROPEAN_PORTUGUESE)

        if len(logging_languages) > MAX_NUMBER_OF_LOGGING_LANGUAGES:
            languages_str = ""
        elif len(logging_languages) > 1:
            languages_str = (
                ", ".join([lang.name for lang in logging_languages[:-1]])
                + f" and {logging_languages[-1].name}"
                + " "
            )
        else:
            languages_str = logging_languages[0].name + " "

        task_str = self.task.name.replace("-", " ")
        dataset_name_str = (
            self.pretty_name or self.name.replace("-", " ").replace("_", " ").title()
        )
        return (
            f"the {truncated_str}{languages_str}{task_str} dataset {dataset_name_str}"
        )

    @property
    def main_language(self) -> Language | tuple[Language, Language]:
        """Get the main language of the dataset.

        Returns:
            The main language or languages of the dataset.

        Raises:
            InvalidBenchmark:
                If the dataset has no languages.
        """
        # Importing here to avoid circular imports
        from .tasks import TRANSLATION

        # Special case for datasets with multiple languages
        if self.task == TRANSLATION:
            return (self.languages[0], self.languages[1])

        match len(self.languages):
            case 0:
                raise InvalidBenchmark(
                    f"Dataset {self.name!r} must have at least one language."
                )
            case 1:
                return self.languages[0]
            case _:
                if ENGLISH in self.languages:
                    return ENGLISH
                elif NORWEGIAN in self.languages:
                    return NORWEGIAN
                elif PORTUGUESE in self.languages:
                    return PORTUGUESE
                else:
                    return self.languages[0]

    @property
    def id2label(self) -> "HashableDict":
        """The mapping from ID to label."""
        return HashableDict({idx: label for idx, label in enumerate(self.labels)})

    @property
    def label2id(self) -> "HashableDict":
        """The mapping from label to ID."""
        return HashableDict({label: i for i, label in enumerate(self.labels)})

    @property
    def num_labels(self) -> int:
        """The number of labels in the dataset."""
        return len(self.labels)

    def __hash__(self) -> int:
        """Return a hash of the dataset configuration."""
        return hash(self.name)

    def get_labels_str(self, labels: c.Sequence[str] | None = None) -> str:
        """Converts a set of labels to a natural string, in the specified language.

        If the task is NER, we separate using 'and' and use the mapped labels instead of
        the BIO NER labels.

        Args:
            labels (optional):
                The labels to convert to a natural string. If None, uses all the labels
                in the dataset. Defaults to None.

        Returns:
            The natural string representation of the labels in specified language.
        """
        if self.task.task_group == TaskGroup.TOKEN_CLASSIFICATION:
            sep_word = self.main_language.and_separator
        else:
            sep_word = self.main_language.or_separator

        if labels is None:
            labels = list()
            for english_label in self.labels:
                if english_label not in self.prompt_label_mapping:
                    continue
                label = self.prompt_label_mapping[english_label]
                if label not in labels:
                    labels.append(label)

        # Convert labels to single-quoted labels - and remove duplicates
        quoted_labels = [f"'{label}'" for label in labels]

        if not quoted_labels:
            return ""
        elif len(quoted_labels) == 1:
            return quoted_labels[0]
        elif len(quoted_labels) == 2:
            return f"{quoted_labels[0]} {sep_word} {quoted_labels[1]}"
        else:
            return f"{', '.join(quoted_labels[:-1])} {sep_word} {quoted_labels[-1]}"


@dataclass
class BenchmarkConfig:
    """General benchmarking configuration, across datasets and models.

    Attributes:
        datasets:
            The datasets to benchmark on.
        finetuning_batch_size:
            The batch size to use for finetuning.
        raise_errors:
            Whether to raise errors instead of skipping them.
        cache_dir:
            Directory to store cached models and datasets.
        api_key:
            The API key to use for a given inference API.
        api_base:
            The base URL for a given inference API. Only relevant if `model` refers to a
            model on an inference API.
        api_version:
            The version of the API to use. Only relevant if `model` refers to a model on
            an inference API.
        progress_bar:
            Whether to show a progress bar.
        save_results:
            Whether to save the benchmark results to 'euroeval_benchmark_results.json'.
        device:
            The device to use for benchmarking.
        trust_remote_code:
            Whether to trust remote code when loading models from the Hugging Face Hub.
        clear_model_cache:
            Whether to clear the model cache after benchmarking each model.
        evaluate_test_split:
            Whether to evaluate on the test split.
        few_shot:
            Whether to only evaluate the model using few-shot evaluation. Only relevant
            if the model is generative.
        num_iterations:
            The number of iterations each model should be evaluated for.
        gpu_memory_utilization:
            The GPU memory utilization to use for vLLM. A larger value will result in
            faster evaluation, but at the risk of running out of GPU memory. Only reduce
            this if you are running out of GPU memory. Only relevant if the model is
            generative.
        attention_backend:
            The attention backend to use for vLLM. Defaults to FLASHINFER. Only
            relevant if the model is generative.
        requires_safetensors:
            Whether to only allow models that use the safetensors format.
        generative_type:
            The type of generative model to benchmark. Only relevant if the model is
            generative.
        download_only:
            Whether to only download the models, metrics and datasets without
            evaluating.
        force:
            Whether to force the benchmark to run even if the results are already
            cached.
        verbose:
            Whether to print verbose output.
        debug:
            Whether to run the benchmark in debug mode.
        run_with_cli:
            Whether the benchmark is being run with the CLI.
    """

    datasets: c.Sequence[DatasetConfig]
    languages: c.Sequence[Language]
    finetuning_batch_size: int
    raise_errors: bool
    cache_dir: str
    api_key: str | None
    api_base: str | None
    api_version: str | None
    progress_bar: bool
    save_results: bool
    device: torch.device
    trust_remote_code: bool
    clear_model_cache: bool
    evaluate_test_split: bool
    few_shot: bool
    num_iterations: int
    gpu_memory_utilization: float
    attention_backend: t.Literal[
        *ATTENTION_BACKENDS  # pyrefly: ignore[invalid-literal]
    ]
    requires_safetensors: bool
    generative_type: GenerativeType | None
    download_only: bool
    force: bool
    verbose: bool
    debug: bool
    run_with_cli: bool

    @property
    def tasks(self) -> c.Sequence[Task]:
        """Get the tasks in the benchmark configuration."""
        return list({dataset_config.task for dataset_config in self.datasets})

    def __post_init__(self) -> None:
        """Post-initialisation checks."""
        # Set dummy API key if it has not been set and we're benchmarking a model on an
        # inference API
        if self.api_key is None and self.api_base is not None:
            self.api_key = "dummy"


class BenchmarkConfigParams(pydantic.BaseModel):
    """The parameters for the benchmark configuration."""

    model_config = pydantic.ConfigDict(
        protected_namespaces=(), arbitrary_types_allowed=True
    )

    task: str | Task | c.Sequence[str | Task] | None
    dataset: str | DatasetConfig | c.Sequence[str | DatasetConfig] | None
    progress_bar: bool
    save_results: bool
    language: str | c.Sequence[str]
    device: Device | None
    finetuning_batch_size: int
    raise_errors: bool
    cache_dir: str
    api_key: str | None
    api_base: str | None
    api_version: str | None
    trust_remote_code: bool
    clear_model_cache: bool
    evaluate_test_split: bool
    few_shot: bool
    num_iterations: int
    requires_safetensors: bool
    download_only: bool
    gpu_memory_utilization: float
    attention_backend: t.Literal[
        *ATTENTION_BACKENDS  # pyrefly: ignore[invalid-literal]
    ]
    generative_type: GenerativeType | None
    custom_datasets_file: Path
    force: bool
    verbose: bool
    debug: bool
    run_with_cli: bool


class BenchmarkResult(pydantic.BaseModel):
    """A benchmark result."""

    dataset: str
    task: str
    languages: c.Sequence[str]
    model: str
    results: ScoreDict
    num_model_parameters: int
    max_sequence_length: int
    vocabulary_size: int
    merge: bool
    generative: bool
    generative_type: str | None
    few_shot: bool | None
    validation_split: bool | None
    euroeval_version: str | None = get_package_version("euroeval")
    transformers_version: str | None = get_package_version("transformers")
    torch_version: str | None = get_package_version("torch")
    vllm_version: str | None = get_package_version("vllm")
    xgrammar_version: str | None = get_package_version("xgrammar")

    @classmethod
    def from_dict(cls, config: dict) -> "BenchmarkResult":
        """Create a benchmark result from a dictionary.

        Args:
            config:
                The configuration dictionary.

        Returns:
            The benchmark result.
        """
        # To be backwards compatible, we accept old results which changed the model
        # name with parameters rather than adding them as explicit parameters
        val_matches = re.search(r"\(.*val.*\)$", config["model"])
        few_shot_matches = re.search(r"\(.*few-shot.*\)$", config["model"])
        zero_shot_matches = re.search(r"\(.*zero-shot.*\)$", config["model"])
        config["model"] = re.sub(
            r"\(.*(few-shot|val).*\)$", "", config["model"]
        ).strip()

        if "merge" not in config:
            config["merge"] = False
        if "generative" not in config:
            config["generative"] = (
                few_shot_matches is not None or zero_shot_matches is not None
            )
        if "generative_type" not in config:
            config["generative_type"] = None
        if "few_shot" not in config:
            config["few_shot"] = zero_shot_matches is None
        if "validation_split" not in config:
            config["validation_split"] = val_matches is not None

        # Backwards compatibility
        if "dataset_languages" in config:
            config["languages"] = config.pop("dataset_languages")

        return cls(**config)

    def append_to_results(self, results_path: Path) -> None:
        """Append the benchmark result to the results file.

        Args:
            results_path:
                The path to the results file.
        """
        json_str = json.dumps(self.model_dump())
        with results_path.open("a") as f:
            f.write("\n" + json_str)


@dataclass
class ModelConfig:
    """Configuration for a model.

    Attributes:
        model_id:
            The ID of the model.
        revision:
            The revision of the model.
        param:
            The parameter of the model, or None if the model has no parameters.
        task:
            The task that the model was trained on.
        languages:
            The languages of the model.
        inference_backend:
            The backend used to perform inference with the model.
        merge:
            Whether the model is a merged model.
        model_type:
            The type of the model (e.g., encoder, base decoder, instruction tuned).
        fresh:
            Whether the model is freshly initialised.
        model_cache_dir:
            The directory to cache the model in.
        adapter_base_model_id:
            The model ID of the base model if the model is an adapter model. Can be None
            if the model is not an adapter model.
        generation_config (optional):
            The generation configuration for generative models, if specified in the
            model repository. Defaults to no generation configuration.
    """

    model_id: str
    revision: str
    param: str | None
    task: str
    languages: c.Sequence[Language]
    inference_backend: "InferenceBackend"
    merge: bool
    model_type: ModelType
    fresh: bool
    model_cache_dir: str
    adapter_base_model_id: str | None
    generation_config: GenerationConfig | None = None

    def __hash__(self) -> int:
        """Return a hash of the model configuration."""
        return hash(self.model_id)


@dataclass
class PreparedModelInputs:
    """The inputs to a model.

    Attributes:
        texts:
            The texts to input to the model. Can be None if the input IDs and attention
            mask are provided instead.
        input_ids:
            The input IDs of the texts. Can be None if the texts are provided instead.
        attention_mask:
            The attention mask of the texts. Can be None if the texts are provided
            instead.
    """

    texts: c.Sequence[str] | None = None
    input_ids: torch.Tensor | None = None
    attention_mask: torch.Tensor | None = None


@dataclass
class GenerativeModelOutput:
    """The output of a generative model.

    Attributes:
        sequences:
            The generated sequences.
        scores:
            The scores of the sequences. This is an array of shape (batch_size,
            num_tokens, num_logprobs, 2), where the last dimension contains the
            token and its logprob. Can be None if the scores are not available.
    """

    sequences: c.Sequence[str]
    scores: c.Sequence[c.Sequence[c.Sequence[tuple[str, float]]]] | None = None


@dataclass
class SingleGenerativeModelOutput:
    """A single output of a generative model.

    Attributes:
        sequence:
            The generated sequence.
        scores:
            The scores of the sequence. This is an array of shape (num_tokens,
            num_logprobs, 2), where the last dimension contains the token and its
            logprob. Can be None if the scores are not available.
    """

    sequence: str
    scores: c.Sequence[c.Sequence[tuple[str, float]]] | None = None


@dataclass
class HFModelInfo:
    """Information about a Hugging Face model.

    Attributes:
        pipeline_tag:
            The pipeline tag of the model.
        tags:
            The other tags of the model.
        adapter_base_model_id:
            The model ID of the base model if the model is an adapter model. Can be None
            if the model is not an adapter model.
    """

    pipeline_tag: str
    tags: c.Sequence[str]
    adapter_base_model_id: str | None


@dataclass
class ModelIdComponents:
    """A model ID split into its components.

    Attributes:
        model_id:
            The main model ID without revision or parameters.
        revision:
            The revision of the model, if any.
        param:
            The parameter of the model, if any.
    """

    model_id: str
    revision: str
    param: str | None


class HashableDict(dict):
    """A hashable dictionary."""

    def __hash__(self) -> int:  # type: ignore[override]
        """Return the hash of the dictionary."""
        return hash(frozenset(self.items()))



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==4.0.0",
#     "huggingface-hub==0.34.4",
# ]
# ///
"""Create the IFEval instruction-following datasets and upload to HF Hub."""

import json

from datasets import DatasetDict, load_dataset
from huggingface_hub import HfApi

LANGUAGES = {
    "ca": "projecte-aina/IFEval_ca",
    "da": "danish-foundation-models/ifeval-da",
    "de": "jzhang86/de_ifeval",
    "it": "mii-llm/ifeval-ita",
    "el": "ilsp/ifeval_greek",
    "en": "tartuNLP/ifeval_en",
    "es": "BSC-LT/IFEval_es",
    "et": "tartuNLP/ifeval_et",
    "fi": "LumiOpen/ifeval_mt::fi",
    "fr": "json:https://raw.githubusercontent.com/lightblue-tech/M-IFEval/refs/heads/main/data/fr_input_data.jsonl",
    "pt": "facebook/Multi-IF?language=Portuguese",
    "sv": "LumiOpen/ifeval_mt::sv",
    "uk": "INSAIT-Institute/ifeval_ukr",
}
TARGET_REPO = "EuroEval/ifeval-{language}"

PROMPT_COLUMN_CANDIDATES = ["prompt", "promptly", "turn_1_prompt"]
INSTRUCTION_ID_LIST_COLUMN_CANDIDATES = [
    "instruction_id_list",
    "categories",
    "turn_1_instruction_id_list",
]
KWARGS_COLUMN_CANDIDATES = ["kwargs", "turn_1_kwargs"]


def main() -> None:
    """Create the IFEval datasets and upload to HF Hub.

    Raises:
        ValueError:
            If the dataset has more than one split, or if the columns could not be
            properly identified.
    """
    for language in LANGUAGES:
        source_repo_id = LANGUAGES[language]

        # Extract the source repo ID, subset and arguments
        if "::" in source_repo_id:
            if "?" in source_repo_id:
                source_repo_id, args = source_repo_id.split("?")
                args, subset = args.split("::")
            else:
                source_repo_id, subset = source_repo_id.split("::")
                args = None
        elif "?" in source_repo_id:
            source_repo_id, args = source_repo_id.split("?")
            subset = None
        else:
            args = None
            subset = None

        if source_repo_id.startswith("json:"):
            source_repo_id = source_repo_id[len("json:") :]
            dataset = load_dataset("json", data_files={"test": source_repo_id})
        else:
            dataset = load_dataset(source_repo_id, name=subset)

        if len(dataset) > 1:
            raise ValueError(
                f"Dataset {source_repo_id} has more than one split. This is currently "
                f"not supported."
            )

        if args is not None:
            filter_dict = {
                arg.split("=")[0]: arg.split("=")[1] for arg in args.split("&")
            }
            dataset = dataset.filter(
                lambda x: all(x[arg] == value for arg, value in filter_dict.items())
            )

        # Ensure that the single split is called "test"
        split_name = list(dataset.keys())[0]
        if split_name != "test":
            dataset = DatasetDict({"test": dataset[split_name]})

        for column in PROMPT_COLUMN_CANDIDATES:
            if column in dataset.column_names["test"]:
                prompt_column = column
                break
        else:
            raise ValueError(f"No prompt column found: {dataset.column_names['test']}")

        for column in INSTRUCTION_ID_LIST_COLUMN_CANDIDATES:
            if column in dataset.column_names["test"]:
                instruction_id_list_column = column
                break
        else:
            raise ValueError(
                f"No instruction_id_list column found: {dataset.column_names['test']}"
            )

        for column in KWARGS_COLUMN_CANDIDATES:
            if column in dataset.column_names["test"]:
                kwargs_column = column
                break
        else:
            raise ValueError(f"No kwargs column found: {dataset.column_names['test']}")

        def transform(row: dict) -> dict:
            """Transform the dataset to match the expected format.

            Args:
                row: The row to transform.

            Returns:
                The transformed row.
            """
            prompt = row[prompt_column]
            instruction_id_list = row[instruction_id_list_column]

            # Ensure that the prompt is a non-JSON string
            try:
                prompt = json.loads(prompt)
            except json.JSONDecodeError:
                # If the prompt is not valid JSON, keep the original string value.
                pass
            if isinstance(prompt, dict):
                prompt = prompt["content"]

            # Ensure that the instruction_id_list is a list of strings
            if isinstance(instruction_id_list, str):
                instruction_id_list = json.loads(instruction_id_list)

            # Ensure that kwargs is a list of dicts
            kwargs = row[kwargs_column]
            if isinstance(kwargs, str):
                kwargs = json.loads(kwargs)
            if any(isinstance(kwarg, str) for kwarg in kwargs):
                kwargs = [json.loads(kwarg) for kwarg in kwargs]
            if isinstance(kwargs, dict):
                kwargs = [kwargs] * len(instruction_id_list)

            return dict(
                text=prompt,
                target_text=dict(
                    instruction_id_list=instruction_id_list, kwargs=kwargs
                ),
            )

        dataset = dataset.map(transform).select_columns(["text", "target_text"])

        target_repo = TARGET_REPO.format(language=language)
        HfApi().delete_repo(repo_id=target_repo, repo_type="dataset", missing_ok=True)
        dataset.push_to_hub(repo_id=target_repo, private=True)


if __name__ == "__main__":
    main()



"""Utility functions for asyncronous tasks."""

import asyncio
import typing as t

from .constants import T


def safe_run(coroutine: t.Coroutine[t.Any, t.Any, T]) -> T:
    """Run a coroutine, ensuring that the event loop is always closed when we're done.

    Args:
        coroutine:
            The coroutine to run.

    Returns:
        The result of the coroutine.
    """
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:  # If the current event loop is closed
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    response = loop.run_until_complete(coroutine)
    return response


async def add_semaphore_and_catch_exception(
    coroutine: t.Coroutine[t.Any, t.Any, T], semaphore: asyncio.Semaphore
) -> T | Exception:
    """Run a coroutine with a semaphore.

    Args:
        coroutine:
            The coroutine to run.
        semaphore:
            The semaphore to use.

    Returns:
        The result of the coroutine.
    """
    async with semaphore:
        try:
            return await coroutine
        except Exception as exc:
            return exc



"""Utility functions to be used in other scripts."""

import gc
import logging
import os
import random
import socket
import sys
import typing as t
from pathlib import Path

import huggingface_hub as hf_hub
import numpy as np
import torch
from huggingface_hub.errors import LocalTokenNotFoundError
from langdetect import DetectorFactory
from requests.exceptions import RequestException

from .caching_utils import cache_arguments
from .constants import LOCAL_MODELS_REQUIRED_FILES
from .exceptions import InvalidModel, NaNValueInModelOutput
from .logging_utils import log_once

if t.TYPE_CHECKING:
    from .types import Predictions


def resolve_model_path(download_dir: str) -> str:
    """Resolve the path to the directory containing the model config files and weights.

    Args:
        download_dir:
            The download directory

    Returns:
        The path to the model.

    Raises:
        InvalidModel:
            If the model path is not valid, or if required files are missing.
    """
    model_path = Path(download_dir)

    # Get the 'path safe' version of the model id, which is the last dir in the path
    model_id_path = model_path.name

    # Hf hub `cache_dir` puts the files in models--`model_id_path`/snapshots
    model_path = model_path / f"models--{model_id_path}" / "snapshots"
    if not model_path.exists():
        raise InvalidModel(
            f"Attempted to load models from the {model_path} directory, "
            "but it does not exist."
        )

    # Get all files in the model path
    found_files = [
        found_file for found_file in model_path.rglob("*") if found_file.is_file()
    ]
    if not found_files:
        raise InvalidModel(f"No model files found at {model_path}")

    # Make sure that there arent multiples of the files found
    if len(found_files) == len(set(found_files)):
        raise InvalidModel(
            f"Found multiple model config files for {model_id_path.strip('models--')}"
            f"at {model_path}"
        )

    # Check that found_files contains at least one of the required files
    found_required_file = next(
        (file for file in found_files if file.name in LOCAL_MODELS_REQUIRED_FILES), None
    )
    if found_required_file is None:
        raise InvalidModel(
            f"At least one of the files {LOCAL_MODELS_REQUIRED_FILES} must be present "
            f"for {model_id_path.strip('models--')} at {model_path}"
        )
    model_path = found_required_file.parent

    # As a precaution we also check that all of the files are in the same directory
    # if not we create a new dir with symlinks to all of the files from all snapshots
    # this is especially useful for vllm where we can only specify one folder and e.g.,
    # the safetensors version of the weights was added in an unmerged PR
    if not all(
        [found_file.parent == found_files[0].parent for found_file in found_files]
    ):
        new_model_path = model_path.parent / "model_files"
        new_model_path.mkdir(exist_ok=True)
        for found_file in found_files:
            Path(new_model_path / found_file.name).symlink_to(found_file)
        model_path = new_model_path

    return str(model_path)


def clear_memory() -> None:
    """Clears the memory of unused items."""
    for gc_generation in range(3):
        gc.collect(generation=gc_generation)
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
    if torch.backends.mps.is_available():
        torch.mps.empty_cache()


def enforce_reproducibility(seed: int = 4242) -> np.random.Generator:
    """Ensures reproducibility of experiments.

    Args:
        seed:
            Seed for the random number generator.

    Returns:
        A numpy random generator
    """
    random.seed(seed)
    np.random.seed(seed)
    rng = np.random.default_rng(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    os.environ["CUDA_LAUNCH_BLOCKING"] = "1"
    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    torch.use_deterministic_algorithms(True, warn_only=True)
    DetectorFactory.seed = 0
    return rng


def get_min_cuda_compute_capability() -> float | None:
    """Gets the lowest cuda capability.

    Returns:
        Device capability as float, or None if CUDA is not available.
    """
    if not torch.cuda.is_available():
        return None

    device_range = range(torch.cuda.device_count())
    capabilities = map(torch.cuda.get_device_capability, device_range)
    major, minor = min(capabilities)
    return float(f"{major}.{minor}")


@cache_arguments(disable_condition=lambda: hasattr(sys, "_called_from_test"))
def internet_connection_available() -> bool:
    """Checks if internet connection is available.

    Returns:
        Whether or not internet connection is available.
    """
    internet_available: bool = False

    try:
        s = socket.create_connection(("1.1.1.1", 80))
        s.close()
        internet_available = True
    except OSError:
        pass
    except Exception as e:
        pytest_socket_errors = ["SocketConnectBlockedError", "SocketBlockedError"]
        if type(e).__name__ not in pytest_socket_errors:
            raise e

    return internet_available


def raise_if_model_output_contains_nan_values(model_output: "Predictions") -> None:
    """Raise an exception if the model output contains NaN values.

    Args:
        model_output:
            The model output to check.

    Raises:
        NaNValueInModelOutput:
            If the model output contains NaN values.
    """
    if isinstance(model_output, np.ndarray):
        if model_output.dtype == np.float32 and np.isnan(model_output).any():
            raise NaNValueInModelOutput()
    elif len(model_output) > 0:
        if isinstance(model_output[0], str):
            if any(x != x for x in model_output):
                raise NaNValueInModelOutput()
        elif len(model_output[0]) > 0:
            if any(x != x for sublist in model_output for x in sublist):
                raise NaNValueInModelOutput()


@cache_arguments()
def get_hf_token(api_key: str | None) -> str | bool:
    """Get the Hugging Face token.

    Args:
        api_key:
            The API key to use as the Hugging Face token. If None, we will try to
            extract it in other ways.

    Returns:
        The Hugging Face token, or True if no token is set but the user is logged in, or
        False if no token is set and the user is not logged in.
    """
    if api_key is not None:
        log_once(
            "Using the Hugging Face API key passed to the function.",
            level=logging.DEBUG,
        )
        return api_key
    elif (token := os.getenv("HF_TOKEN")) is not None:
        log_once(
            "Using the Hugging Face API key from the environment variable `HF_TOKEN`.",
            level=logging.DEBUG,
        )
        return token
    try:
        hf_hub.whoami()
        log_once(
            "No Hugging Face API key was set, but the user is logged in to Hugging "
            "Face, so using the local token.",
            level=logging.DEBUG,
        )
        return True
    except LocalTokenNotFoundError:
        log_once(
            "No Hugging Face API key was set and the user is not logged in to Hugging "
            "Face, so no token will be used.",
            level=logging.DEBUG,
        )
        return False
    except RequestException:
        log_once(
            "No Hugging Face API key was set and the connection to Hugging Face "
            "failed, so no token will be used.",
            level=logging.DEBUG,
        )
        return False



"""All dataset configurations used in EuroEval."""

import collections.abc as c
import logging
from pathlib import Path

from ..custom_dataset_configs import (
    load_custom_datasets_module,
    try_get_dataset_config_from_repo,
)
from ..data_models import DatasetConfig
from ..languages import get_all_languages
from ..logging_utils import log_once
from ..tasks import SPEED
from .albanian import *  # noqa: F403
from .bosnian import *  # noqa: F403
from .bulgarian import *  # noqa: F403
from .catalan import *  # noqa: F403
from .croatian import *  # noqa: F403
from .czech import *  # noqa: F403
from .danish import *  # noqa: F403
from .dutch import *  # noqa: F403
from .english import *  # noqa: F403
from .estonian import *  # noqa: F403
from .faroese import *  # noqa: F403
from .finnish import *  # noqa: F403
from .french import *  # noqa: F403
from .german import *  # noqa: F403
from .greek import *  # noqa: F403
from .hungarian import *  # noqa: F403
from .icelandic import *  # noqa: F403
from .italian import *  # noqa: F403
from .latvian import *  # noqa: F403
from .lithuanian import *  # noqa: F403
from .norwegian import *  # noqa: F403
from .polish import *  # noqa: F403
from .portuguese import *  # noqa: F403
from .romanian import *  # noqa: F403
from .serbian import *  # noqa: F403
from .slovak import *  # noqa: F403
from .slovene import *  # noqa: F403
from .spanish import *  # noqa: F403
from .swedish import *  # noqa: F403
from .translation_configs import *  # noqa: F403
from .ukrainian import *  # noqa: F403


def get_all_dataset_configs(
    custom_datasets_file: Path,
    dataset_ids: c.Sequence[str],
    api_key: str | None,
    cache_dir: Path,
) -> dict[str, DatasetConfig]:
    """Get a mapping of all the dataset configurations.

    Args:
        custom_datasets_file:
            A path to a Python file containing custom dataset configurations.
        dataset_ids:
            The IDs of the datasets to include in the mapping.
        api_key:
            The Hugging Face API key to use to check if the repositories have custom
            dataset configs.
        cache_dir:
            The directory to store the cache in.

    Returns:
        A mapping between names of datasets and their configurations.
    """
    globals_dict = globals()

    # If any of the dataset IDs are referring to Hugging Face dataset IDs, then we check
    # if the repositories have custom dataset configs and if they do, we add them to the
    # globals dict.
    for dataset_id in dataset_ids:
        dataset_config_or_none = try_get_dataset_config_from_repo(
            dataset_id=dataset_id, api_key=api_key, cache_dir=cache_dir
        )
        if dataset_config_or_none is not None:
            globals_dict[dataset_id] = dataset_config_or_none
            msg = f"Loaded external dataset {dataset_id}"
            split_strings = []
            if dataset_config_or_none.train_split is not None:
                split_strings.append(
                    f"train split '{dataset_config_or_none.train_split}'"
                )
            if dataset_config_or_none.val_split is not None:
                split_strings.append(f"val split '{dataset_config_or_none.val_split}'")
            if dataset_config_or_none.test_split is not None:
                split_strings.append(
                    f"test split '{dataset_config_or_none.test_split}'"
                )
            if split_strings:
                msg += f" with {', '.join(split_strings[:-1])} and {split_strings[-1]}"
            msg += "."
            log_once(msg, level=logging.INFO)

    # Add the custom datasets from the custom datasets file to the globals dict
    module = load_custom_datasets_module(custom_datasets_file=custom_datasets_file)
    if module is not None:
        globals_dict |= vars(module)

    # Extract the dataset configs from the globals dict
    dataset_configs = [
        cfg
        for cfg in globals_dict.values()
        if isinstance(cfg, DatasetConfig) and cfg.task != SPEED
    ]
    assert len(dataset_configs) == len({cfg.name for cfg in dataset_configs}), (
        "There are duplicate dataset configurations. Please ensure that each dataset "
        "has a unique name."
    )

    mapping = {cfg.name: cfg for cfg in dataset_configs}
    return mapping


SPEED_CONFIG = DatasetConfig(
    name="speed",
    pretty_name="",
    source="",
    task=SPEED,
    languages=list(get_all_languages().values()),
)



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==3.5.0",
#     "huggingface-hub==0.24.0",
#     "pandas==2.2.0",
#     "requests==2.32.3",
# ]
# ///

"""Create the Belarusian Sentiment dataset and upload it to the HF Hub."""

import pandas as pd
from datasets import Dataset, DatasetDict, Split, load_dataset
from huggingface_hub import HfApi

from .constants import MAX_NUM_CHARS_IN_DOCUMENT, MIN_NUM_CHARS_IN_DOCUMENT  # noqa


def main() -> None:
    """Create the Belarusian Sentiment dataset and upload it to the HF Hub."""
    # Define the base download URL
    repo_id = "maaxap/BelarusianGLUE"

    # Download the dataset
    dataset = load_dataset(path=repo_id, name="besls", token=True)
    assert isinstance(dataset, DatasetDict)

    # Convert the dataset to dataframes
    train_df = dataset["train"].to_pandas()
    val_df = dataset["validation"].to_pandas()
    test_df = dataset["test"].to_pandas()
    assert isinstance(train_df, pd.DataFrame)
    assert isinstance(val_df, pd.DataFrame)
    assert isinstance(test_df, pd.DataFrame)

    train_df = process(df=train_df)
    val_df = process(df=val_df)
    test_df = process(df=test_df)

    final_train_df, final_val_df, final_test_df = make_splits(
        df_train=train_df, df_val=val_df, df_test=test_df
    )

    # Create DatasetDict
    dataset = DatasetDict(
        {
            "train": Dataset.from_pandas(final_train_df, split=Split.TRAIN),
            "val": Dataset.from_pandas(final_val_df, split=Split.VALIDATION),
            "test": Dataset.from_pandas(final_test_df, split=Split.TEST),
        }
    )

    # Create dataset ID
    dataset_id = "EuroEval/besls"

    # Remove the dataset from Hugging Face Hub if it already exists
    HfApi().delete_repo(dataset_id, repo_type="dataset", missing_ok=True)

    # Push the dataset to the Hugging Face Hub
    dataset.push_to_hub(dataset_id, private=True)


def process(df: pd.DataFrame) -> pd.DataFrame:
    """Process the dataframe.

    Args:
        df: The dataframe to process.

    Returns:
        The processed dataframe.
    """
    # Rename sentence to text
    df.rename(columns={"sentence": "text"}, inplace=True)

    # Keep only relevant columns
    columns_to_keep = ["text", "label"]
    df = df[columns_to_keep]

    # Map labels
    label_mapping = {0: "negative", 1: "positive"}
    df["label"] = df["label"].map(lambda x: label_mapping[int(x)])

    # Filter by text length
    df = _filter_by_length(df)
    return df


def make_splits(
    df_train: pd.DataFrame, df_val: pd.DataFrame, df_test: pd.DataFrame
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Make splits for the dataset.

    Args:
        df_train: The training dataframe.
        df_val: The validation dataframe.
        df_test: The test dataframe.

    Returns:
        The final training, validation, and test dataframes.
    """
    train_size = 256
    val_size = 128

    # Sample train and val split and keep remaining samples for test
    final_train_df = df_train.sample(n=train_size, random_state=4242)
    remaining_train_df = df_train[~df_train.index.isin(final_train_df.index)]

    final_val_df = df_val.sample(n=val_size, random_state=4242)
    remaining_val_df = df_val[~df_val.index.isin(final_val_df.index)]

    final_test_df = pd.concat(
        [remaining_train_df, remaining_val_df, df_test], ignore_index=True
    )

    # Reset indices
    final_train_df = final_train_df.reset_index(drop=True)
    final_val_df = final_val_df.reset_index(drop=True)
    final_test_df = final_test_df.reset_index(drop=True)

    assert isinstance(final_train_df, pd.DataFrame)
    assert isinstance(final_val_df, pd.DataFrame)
    assert isinstance(final_test_df, pd.DataFrame)

    return final_train_df, final_val_df, final_test_df


def _filter_by_length(df: pd.DataFrame) -> pd.DataFrame:
    """Filter dataframe by text length.

    Args:
        df:
            The dataframe to filter.

    Returns:
        The filtered dataframe.
    """
    new_df = df.copy()
    new_df["text_len"] = new_df.text.str.len()
    return new_df.query("text_len >= @MIN_NUM_CHARS_IN_DOCUMENT").query(
        "text_len <= @MAX_NUM_CHARS_IN_DOCUMENT"
    )


if __name__ == "__main__":
    main()



# /// script
# requires-python = ">=3.10,<4.0"
# dependencies = [
#     "datasets==4.0.0",
#     "huggingface-hub==0.34.4",
#     "tqdm==4.67.1",
# ]
# ///

"""Create the WMT24++ datasets."""

from datasets import Dataset, DatasetDict, load_dataset
from tqdm.auto import tqdm

LANGUAGES = {
    "bg": "en-bg_BG",
    "ca": "en-ca_ES",
    "cs": "en-cs_CZ",
    "da": "en-da_DK",
    "de": "en-de_DE",
    "el": "en-el_GR",
    "et": "en-et_EE",
    "fi": "en-fi_FI",
    "fr": "en-fr_FR",
    "hr": "en-hr_HR",
    "hu": "en-hu_HU",
    "is": "en-is_IS",
    "it": "en-it_IT",
    "lt": "en-lt_LT",
    "lv": "en-lv_LV",
    "nl": "en-nl_NL",
    "no": "en-no_NO",
    "pl": "en-pl_PL",
    "pt": "en-pt_PT",
    "ro": "en-ro_RO",
    "sk": "en-sk_SK",
    "sl": "en-sl_SI",
    "sr": "en-sr_RS",
    "sv": "en-sv_SE",
    "uk": "en-uk_UA",
}


def main() -> None:
    """Create the WMT24++ datasets."""
    target_repo_id = "EuroEval/wmt24pp-en-{target}"

    for language, subset in tqdm(
        iterable=LANGUAGES.items(), desc="Creating WMT24++ datasets", unit="language"
    ):
        ds = load_dataset("google/wmt24pp", name=subset, split="train").shuffle(seed=42)
        assert isinstance(ds, Dataset)

        ds = ds.filter(lambda x: not x["is_bad_source"])
        ds = ds.rename_columns(dict(source="text", target="target_text"))

        train_size = 64
        val_size = 128
        test_size = len(ds) - train_size - val_size

        train = ds.select(range(train_size))
        val = ds.select(range(train_size, train_size + val_size))
        test = ds.select(
            range(train_size + val_size, train_size + val_size + test_size)
        )

        assert isinstance(train, Dataset)
        assert isinstance(val, Dataset)
        assert isinstance(test, Dataset)

        new_ds = DatasetDict({"train": train, "val": val, "test": test})
        new_ds.push_to_hub(target_repo_id.format(target=language), private=True)


if __name__ == "__main__":
    main()



"""Dataset configurations for the translation task."""

from ..data_models import DatasetConfig
from ..languages import (
    BULGARIAN,
    CATALAN,
    CROATIAN,
    CZECH,
    DANISH,
    DUTCH,
    ENGLISH,
    ESTONIAN,
    FINNISH,
    FRENCH,
    GERMAN,
    GREEK,
    HUNGARIAN,
    ICELANDIC,
    ITALIAN,
    LATVIAN,
    LITHUANIAN,
    NORWEGIAN,
    POLISH,
    PORTUGUESE,
    ROMANIAN,
    SERBIAN,
    SLOVAK,
    SLOVENE,
    SWEDISH,
    UKRAINIAN,
)
from ..tasks import TRANSLATION

WMT24PP_EN_BG_CONFIG = DatasetConfig(
    name="wmt24pp-en-bg",
    pretty_name="WMT24++-bg",
    source="EuroEval/wmt24pp-en-bg",
    task=TRANSLATION,
    languages=[ENGLISH, BULGARIAN],
    unofficial=True,
)

WMT24PP_EN_CA_CONFIG = DatasetConfig(
    name="wmt24pp-en-ca",
    pretty_name="WMT24++-ca",
    source="EuroEval/wmt24pp-en-ca",
    task=TRANSLATION,
    languages=[ENGLISH, CATALAN],
    unofficial=True,
)

WMT24PP_EN_CS_CONFIG = DatasetConfig(
    name="wmt24pp-en-cs",
    pretty_name="WMT24++-cs",
    source="EuroEval/wmt24pp-en-cs",
    task=TRANSLATION,
    languages=[ENGLISH, CZECH],
    unofficial=True,
)

WMT24PP_EN_DA_CONFIG = DatasetConfig(
    name="wmt24pp-en-da",
    pretty_name="WMT24++-da",
    source="EuroEval/wmt24pp-en-da",
    task=TRANSLATION,
    languages=[ENGLISH, DANISH],
    unofficial=True,
)

WMT24PP_EN_DE_CONFIG = DatasetConfig(
    name="wmt24pp-en-de",
    pretty_name="WMT24++-de",
    source="EuroEval/wmt24pp-en-de",
    task=TRANSLATION,
    languages=[ENGLISH, GERMAN],
    unofficial=True,
)

WMT24PP_EN_EL_CONFIG = DatasetConfig(
    name="wmt24pp-en-el",
    pretty_name="WMT24++-el",
    source="EuroEval/wmt24pp-en-el",
    task=TRANSLATION,
    languages=[ENGLISH, GREEK],
    unofficial=True,
)

WMT24PP_EN_ET_CONFIG = DatasetConfig(
    name="wmt24pp-en-et",
    pretty_name="WMT24++-et",
    source="EuroEval/wmt24pp-en-et",
    task=TRANSLATION,
    languages=[ENGLISH, ESTONIAN],
    unofficial=True,
)

WMT24PP_EN_FI_CONFIG = DatasetConfig(
    name="wmt24pp-en-fi",
    pretty_name="WMT24++-fi",
    source="EuroEval/wmt24pp-en-fi",
    task=TRANSLATION,
    languages=[ENGLISH, FINNISH],
    unofficial=True,
)

WMT24PP_EN_FR_CONFIG = DatasetConfig(
    name="wmt24pp-en-fr",
    pretty_name="WMT24++-fr",
    source="EuroEval/wmt24pp-en-fr",
    task=TRANSLATION,
    languages=[ENGLISH, FRENCH],
    unofficial=True,
)

WMT24PP_EN_HR_CONFIG = DatasetConfig(
    name="wmt24pp-en-hr",
    pretty_name="WMT24++-hr",
    source="EuroEval/wmt24pp-en-hr",
    task=TRANSLATION,
    languages=[ENGLISH, CROATIAN],
    unofficial=True,
)

WMT24PP_EN_HU_CONFIG = DatasetConfig(
    name="wmt24pp-en-hu",
    pretty_name="WMT24++-hu",
    source="EuroEval/wmt24pp-en-hu",
    task=TRANSLATION,
    languages=[ENGLISH, HUNGARIAN],
    unofficial=True,
)

WMT24PP_EN_IS_CONFIG = DatasetConfig(
    name="wmt24pp-en-is",
    pretty_name="WMT24++-is",
    source="EuroEval/wmt24pp-en-is",
    task=TRANSLATION,
    languages=[ENGLISH, ICELANDIC],
    unofficial=True,
)

WMT24PP_EN_IT_CONFIG = DatasetConfig(
    name="wmt24pp-en-it",
    pretty_name="WMT24++-it",
    source="EuroEval/wmt24pp-en-it",
    task=TRANSLATION,
    languages=[ENGLISH, ITALIAN],
    unofficial=True,
)

WMT24PP_EN_LT_CONFIG = DatasetConfig(
    name="wmt24pp-en-lt",
    pretty_name="WMT24++-lt",
    source="EuroEval/wmt24pp-en-lt",
    task=TRANSLATION,
    languages=[ENGLISH, LITHUANIAN],
    unofficial=True,
)

WMT24PP_EN_LV_CONFIG = DatasetConfig(
    name="wmt24pp-en-lv",
    pretty_name="WMT24++-lv",
    source="EuroEval/wmt24pp-en-lv",
    task=TRANSLATION,
    languages=[ENGLISH, LATVIAN],
    unofficial=True,
)

WMT24PP_EN_NL_CONFIG = DatasetConfig(
    name="wmt24pp-en-nl",
    pretty_name="WMT24++-nl",
    source="EuroEval/wmt24pp-en-nl",
    task=TRANSLATION,
    languages=[ENGLISH, DUTCH],
    unofficial=True,
)

WMT24PP_EN_NO_CONFIG = DatasetConfig(
    name="wmt24pp-en-no",
    pretty_name="WMT24++-no",
    source="EuroEval/wmt24pp-en-no",
    task=TRANSLATION,
    languages=[ENGLISH, NORWEGIAN],
    unofficial=True,
)

WMT24PP_EN_PL_CONFIG = DatasetConfig(
    name="wmt24pp-en-pl",
    pretty_name="WMT24++-pl",
    source="EuroEval/wmt24pp-en-pl",
    task=TRANSLATION,
    languages=[ENGLISH, POLISH],
    unofficial=True,
)

WMT24PP_EN_PT_CONFIG = DatasetConfig(
    name="wmt24pp-en-pt",
    pretty_name="WMT24++-pt",
    source="EuroEval/wmt24pp-en-pt",
    task=TRANSLATION,
    languages=[ENGLISH, PORTUGUESE],
    unofficial=True,
)

WMT24PP_EN_RO_CONFIG = DatasetConfig(
    name="wmt24pp-en-ro",
    pretty_name="WMT24++-ro",
    source="EuroEval/wmt24pp-en-ro",
    task=TRANSLATION,
    languages=[ENGLISH, ROMANIAN],
    unofficial=True,
)

WMT24PP_EN_SK_CONFIG = DatasetConfig(
    name="wmt24pp-en-sk",
    pretty_name="WMT24++-sk",
    source="EuroEval/wmt24pp-en-sk",
    task=TRANSLATION,
    languages=[ENGLISH, SLOVAK],
    unofficial=True,
)

WMT24PP_EN_SL_CONFIG = DatasetConfig(
    name="wmt24pp-en-sl",
    pretty_name="WMT24++-sl",
    source="EuroEval/wmt24pp-en-sl",
    task=TRANSLATION,
    languages=[ENGLISH, SLOVENE],
    unofficial=True,
)

WMT24PP_EN_SR_CONFIG = DatasetConfig(
    name="wmt24pp-en-sr",
    pretty_name="WMT24++-sr",
    source="EuroEval/wmt24pp-en-sr",
    task=TRANSLATION,
    languages=[ENGLISH, SERBIAN],
    unofficial=True,
)

WMT24PP_EN_SV_CONFIG = DatasetConfig(
    name="wmt24pp-en-sv",
    pretty_name="WMT24++-sv",
    source="EuroEval/wmt24pp-en-sv",
    task=TRANSLATION,
    languages=[ENGLISH, SWEDISH],
    unofficial=True,
)

WMT24PP_EN_UK_CONFIG = DatasetConfig(
    name="wmt24pp-en-uk",
    pretty_name="WMT24++-uk",
    source="EuroEval/wmt24pp-en-uk",
    task=TRANSLATION,
    languages=[ENGLISH, UKRAINIAN],
    unofficial=True,
)



"""All Belarusian dataset configurations used in EuroEval."""

from ..data_models import DatasetConfig
from ..languages import BELARUSIAN
from ..tasks import COMMON_SENSE, LA, NER, RC, SENT

# Official datasets ###

BESLS_CONFIG = DatasetConfig(
    name="besls",
    pretty_name="BeSLS",
    source="EuroEval/besls",
    task=SENT,
    languages=[BELARUSIAN],
)

SCALA_BE_CONFIG = DatasetConfig(
    name="scala-be",
    pretty_name="ScaLA-be",
    source="EuroEval/scala-be",
    task=LA,
    languages=[BELARUSIAN],
)

WIKIANN_BE_CONFIG = DatasetConfig(
    name="wikiann-be",
    pretty_name="WikiANN-be",
    source="EuroEval/wikiann-be-mini",
    task=NER,
    languages=[BELARUSIAN],
)

MULTI_WIKI_QA_BE_CONFIG = DatasetConfig(
    name="multi-wiki-qa-be",
    pretty_name="MultiWikiQA-be",
    source="EuroEval/multi-wiki-qa-be-mini",
    task=RC,
    languages=[BELARUSIAN],
)

BE_WSC_CONFIG = DatasetConfig(
    name="be-wsc",
    pretty_name="BE-WSC",
    source="EuroEval/be-wsc",
    task=COMMON_SENSE,
    languages=[BELARUSIAN],
)
